/*
 * Copyright 2011 E-Government Innovation Center (EGIZ)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.gv.zustellung.zusemsg.wsdl.client;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URL;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.Security;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.activation.DataHandler;
import javax.annotation.Resource;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;
import javax.net.ssl.SSLContext;
import javax.xml.namespace.QName;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.Handler;
import javax.xml.ws.handler.MessageContext;

import org.apache.commons.codec.binary.Base64InputStream;
import org.apache.commons.lang.RandomStringUtils;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.mail.smime.SMIMEEnvelopedGenerator;
import org.bouncycastle.mail.smime.SMIMEException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import at.gv.util.LaxHostNameVerifier;
import at.gv.util.LoggingHandler;
import at.gv.util.ex.EgovUtilException;
import at.gv.zustellung.zusemsg.ZuseMsgException;
import at.gv.zustellung.zusemsg.config.ZuseMsgConfiguration;
import at.gv.zustellung.zusemsg.wsdl.App2ZusePortType;
import at.gv.zustellung.zusemsg.wsdl.App2ZuseService;
import at.gv.zustellung.zusemsg.wsdl.client.attachments.BinaryAttachment;
import at.gv.zustellung.zusemsg.wsdl.client.attachments.MailBody;
import at.gv.zustellung.zusemsg.xsd.zuse.DeliveryRequestStatusType;
import at.gv.zustellung.zusemsg.xsd.zuse.DeliveryRequestType;

import com.sun.xml.ws.developer.JAXWSProperties;

public class App2ZuseClient {

	@Resource
  WebServiceContext wsContext;

	private ZuseMsgConfiguration config = null;
	private App2ZusePortType port = null;
	Logger log = LoggerFactory.getLogger(App2ZuseClient.class);
	
	static {
		// adding bouncycastle provider
		Security.addProvider(new BouncyCastleProvider());
	}
	
	public App2ZuseClient(ZuseMsgConfiguration config) {
		if (config == null) {
			throw new NullPointerException("Parameter config must not be null.");
		}
		this.config = config;
		
		URL url = App2ZuseClient.class.getResource("/wsdl/app2zuse.wsdl");
		App2ZuseService service = new App2ZuseService(url, new QName("http://reference.e-government.gv.at/namespace/zustellung/app2zuse#wsdl", "app2zuseService"));
		this.port = service.getApp2ZusePortType();

	}
	
  @SuppressWarnings({ "rawtypes" })
  public DeliveryRequestStatusType sendApp2ZuseRequest(String deliveryServerURL,
  		DeliveryRequestType request,
  		MailBody mailBody,
  		List<BinaryAttachment> attachments,
  		boolean makeZusContainer,
  		X509Certificate encryptionCertificate) throws App2ZuseClientException {
    
  	// verify that parameters are not null
  	if (deliveryServerURL == null) {
  		throw new NullPointerException("Parameter 'deliveryServerURL' must not be null.");
  	}
  	log.trace("Delivery Server request URL: " + deliveryServerURL);
  	if (request == null) {
  		throw new NullPointerException("Parameter 'request' must not be null.");
  	}
  	if (mailBody == null) {
  		throw new NullPointerException("Parameter 'mailBody' must not be null.");
  	}
  	if (attachments == null || attachments.size() < 1) {
  		throw new App2ZuseClientException("No attachments found in parameter 'attachments'");
  	}
  	log.trace("Number of attachments: " + attachments.size());
  	
    try {
    
	    // set endpoint address
  		BindingProvider bindingProvider = (BindingProvider) this.port;
  		Map<String, Object> requestContext = bindingProvider.getRequestContext();
    	requestContext.put(
          BindingProvider.ENDPOINT_ADDRESS_PROPERTY, deliveryServerURL);
      
    	// add trace log
    	log.trace("Adding JAX-WS request/response trace handler.");
    	List<Handler> handlerList = bindingProvider.getBinding().getHandlerChain();
    	if (handlerList == null) {
	      handlerList = new ArrayList<Handler>();
      }
	    LoggingHandler loggingHandler = new LoggingHandler();
	    handlerList.add(loggingHandler);
	    bindingProvider.getBinding().setHandlerChain(handlerList);

      // initialize ssl
      if (deliveryServerURL.toLowerCase().startsWith("https")) {
      	log.trace("Using ssl for app2zuse request.");
      	SSLContext sslContext = config.getSSLConfiguration().getSSLContext(false);
      	if (sslContext == null) {
					throw new App2ZuseClientException("SSL context from configuration is empty. Please configure an SSL context in the configuration first.");
				}
				requestContext.put(JAXWSProperties.SSL_SOCKET_FACTORY, sslContext.getSocketFactory());
				
				// check for lax hostname
	      if (this.config.getSSLConfiguration().useLaxHostNameVerifier()) {
	      	log.trace("LaxHostnameVerifier enabled. This setting is not recommended to use.");
	      	requestContext.put(JAXWSProperties.HOSTNAME_VERIFIER, new LaxHostNameVerifier());
	      }
      }
      log.trace("Preparing SwA attachments.");
      // prepare attachments
      Map<String, DataHandler> atts = new HashMap<String, DataHandler>();
      
      log.trace("Preparing message/rfc822 multipart.");
      Properties props = System.getProperties();
      Session session = Session.getDefaultInstance(props, null);
      
      MimeMultipart mmp = new MimeMultipart();
			
      log.trace("Adding mailbody part.");
      // first add mailbody (content)
      MimeBodyPart mbp = new MimeBodyPart();
      mbp.setDisposition("inline");
      mbp.setDataHandler(new DataHandler(mailBody));
      mmp.addBodyPart(mbp);
      
      // handle ZUS ZIP container request
      if (makeZusContainer) {
      	log.debug("Making ZUS container.");
      	// create new mime container
      	MimeMultipart zmp = new MimeMultipart();
      	for (BinaryAttachment attachment : attachments) {
      		log.trace("Attaching " + attachment.getName() + " - Content-type: " + attachment.getContentType());
	      	MimeBodyPart attachmentPart = new MimeBodyPart();
	      	attachmentPart.setFileName(attachment.getName());
	      	attachmentPart.setDisposition("inline");
	      	DataHandler dataHandler = new DataHandler(attachment);
	      	attachmentPart.setDataHandler(dataHandler);
					attachmentPart.setHeader("Content-length", String.valueOf(attachment.getSize()));	      	
	      	zmp.addBodyPart(attachmentPart);
      	}
      	// update mime headers
      	MimeMessage message = new MimeMessage(session);
      	message.setContent(zmp);
        message.saveChanges();
        log.trace("Compressing ZUS container.");
        // make zip
        ByteArrayOutputStream zusBos = new ByteArrayOutputStream();
        ZipOutputStream zos = new ZipOutputStream(zusBos);
        String zusEntryName = config.getZusEntryName();
        if (zusEntryName == null) {
        	zusEntryName = "zuse-content.mime";
        }
        log.trace("ZUS file entry name");
        ZipEntry entry = new ZipEntry(zusEntryName);
        zos.putNextEntry(entry);
        message.writeTo(zos);
        zos.flush();
        zos.close();
        
        log.trace("Adding ZUS body part.");
        MimeBodyPart zusPart = new MimeBodyPart();
        String zusContainerName = config.getZusContainerName();
        if (zusContainerName == null) {
        	zusContainerName = "zuse.zus";
        }
        log.trace("ZUS container name: " + zusContainerName);
        BinaryAttachment zusAttachment = new BinaryAttachment(zusContainerName, "application/x-zus", new Base64InputStream(new ByteArrayInputStream(zusBos.toByteArray()), true));
        zusPart.setDataHandler(new DataHandler(zusAttachment));
        zusPart.setFileName(zusContainerName);
        zusPart.setDisposition("inline");
        mmp.addBodyPart(zusPart);
      } else {
      	log.debug("Making standard container");
	      for (BinaryAttachment attachment : attachments) {
	      	log.trace("Attaching " + attachment.getName() + " - Content-type: " + attachment.getContentType());
	      	MimeBodyPart attachmentPart = new MimeBodyPart();
	      	attachmentPart.setFileName(attachment.getName());
	      	attachmentPart.setDisposition("attachment");
	      	DataHandler dataHandler = new DataHandler(attachment);
	      	attachmentPart.setDataHandler(dataHandler);
	      	mmp.addBodyPart(attachmentPart);
	      }
      }
      
      log.trace("Creating second SwA part.");
    	MimeMessage message = new MimeMessage(session);
    	ByteArrayOutputStream bos = new ByteArrayOutputStream();
    	message.setContent(mmp);
    	message.saveChanges();
    	mmp.setPreamble("This is a multi-part message in MIME format.");
      message.writeTo(bos);

      // encryption enabled
      if (encryptionCertificate != null) {
      	log.debug("Encypting message.");
      	log.trace("Encryption certificate: " + encryptionCertificate.toString());
      	SMIMEEnvelopedGenerator smimeGenerator = new SMIMEEnvelopedGenerator();
      	smimeGenerator.addKeyTransRecipient(encryptionCertificate);
      	MimeBodyPart smimePart = smimeGenerator.generate(message, SMIMEEnvelopedGenerator.AES256_CBC, "BC");
      	bos.reset();
      	smimePart.writeTo(bos);
      	
      	MimeMessage sMimeMessage = new MimeMessage(session, new ByteArrayInputStream(bos.toByteArray()));
      	sMimeMessage.saveChanges();
      	bos.reset();
      	sMimeMessage.writeTo(bos);
      }
      
      // create content-id
      String contentId = "zuse-attachments." + RandomStringUtils.random(10, true, true) + 
      			"." + RandomStringUtils.random(10, true, true) + "@zustellung.gv.at";
      
      BinaryAttachment mimeMessage = new BinaryAttachment(contentId, "message/rfc822", new ByteArrayInputStream(bos.toByteArray()));
      DataHandler rootPartHandler = new DataHandler(mimeMessage);
      atts.put(contentId, rootPartHandler);

      // adding attachments to request context
      requestContext.put(MessageContext.OUTBOUND_MESSAGE_ATTACHMENTS, atts);
      
			return this.port.app2ZuseOperation(request);
      
    } catch (MessagingException e) {
    	throw new App2ZuseClientException(e);
    } catch (ZuseMsgException e) {
    	throw new App2ZuseClientException(e);
    } catch (IOException e) {
    	throw new App2ZuseClientException(e);
    } catch (NoSuchAlgorithmException e) {
    	throw new App2ZuseClientException(e);
    } catch (NoSuchProviderException e) {
    	throw new App2ZuseClientException(e);
    } catch (SMIMEException e) {
    	throw new App2ZuseClientException(e);
    } catch (EgovUtilException e) {
    	throw new App2ZuseClientException(e);
    }
  }
  
}
