/*
 * Copyright 2011 E-Government Innovation Center (EGIZ)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.gv.zustellung.zusemsg.wsdl.client;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import javax.activation.DataHandler;
import javax.annotation.Resource;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Multipart;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.AddressException;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeBodyPart;
import javax.mail.internet.MimeMessage;
import javax.mail.internet.MimeMultipart;
import javax.net.ssl.SSLContext;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.namespace.QName;
import javax.xml.transform.TransformerException;
import javax.xml.ws.BindingProvider;
import javax.xml.ws.WebServiceContext;
import javax.xml.ws.handler.Handler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Element;

import at.gv.util.DOMUtils;
import at.gv.util.LaxHostNameVerifier;
import at.gv.util.LoggingHandler;
import at.gv.util.MiscUtil;
import at.gv.zustellung.zusemsg.ZuseMsgException;
import at.gv.zustellung.zusemsg.config.Zuse2AppConfiguration;
import at.gv.zustellung.zusemsg.config.ZuseMsgConfiguration;
import at.gv.zustellung.zusemsg.wsdl.NotificationPortType;
import at.gv.zustellung.zusemsg.wsdl.Zuse2App;
import at.gv.zustellung.zusemsg.wsdl.client.attachments.BinaryAttachment;
import at.gv.zustellung.zusemsg.wsdl.client.attachments.MailBody;
import at.gv.zustellung.zusemsg.wsdl.client.handler.Zuse2AppHandlerResolver;
import at.gv.zustellung.zusemsg.wsdl.client.handler.Zuse2AppSignatureCreator;
import at.gv.zustellung.zusemsg.wsdl.client.handler.Zuse2AppSignatureException;
import at.gv.zustellung.zusemsg.xsd.zuse.DeliveryNotification;

import com.sun.xml.ws.developer.JAXWSProperties;

public class Zuse2AppClient {

	@Resource
  WebServiceContext wsContext;

	private ZuseMsgConfiguration config = null;
	private NotificationPortType port = null;
	Logger log = LoggerFactory.getLogger(Zuse2AppClient.class);
	private boolean createSignature = true;
	
	public Zuse2AppClient(ZuseMsgConfiguration config, boolean createSignature) {
		if (config == null) {
			throw new NullPointerException("Parameter config must not be null.");
		}
		this.config = config;
		
		URL url = Zuse2AppClient.class.getResource("/wsdl/zuse2app.wsdl");
		Zuse2App service = new Zuse2App(url, new QName("http://reference.e-government.gv.at/namespace/zuse11/zuse2app.wsdl", "zuse2app"));
		this.createSignature = createSignature;
		if (this.createSignature) {
			service.setHandlerResolver(new Zuse2AppHandlerResolver(this.config));
		}
		this.port = service.getNotification();
		
	}
	
  @SuppressWarnings("rawtypes")
  public void sendDeliveryRequestWebService(String webServiceURL, DeliveryNotification request) throws Zuse2AppClientException {
    log.debug("Sending zuse2app notification to application (webservice).");
  	// verify that parameters are not null
  	if (webServiceURL == null) {
  		throw new NullPointerException("Parameter 'deliveryServerURL' must not be null.");
  	}
  	log.trace("Webservice-URL: " + webServiceURL);
  	if (request == null) {
  		throw new NullPointerException("Parameter 'request' must not be null.");
  	}
  	try {
	    // set endpoint address
  		BindingProvider bindingProvider = (BindingProvider) this.port;
    	bindingProvider.getRequestContext().put(
          BindingProvider.ENDPOINT_ADDRESS_PROPERTY, webServiceURL);
      
    	// add trace log
    	log.trace("Adding JAX-WS request/response trace handler.");
    	List<Handler> handlerList = bindingProvider.getBinding().getHandlerChain();
    	if (handlerList == null) {
	      handlerList = new ArrayList<Handler>();
      }
	    LoggingHandler loggingHandler = new LoggingHandler();
	    handlerList.add(loggingHandler);
	    bindingProvider.getBinding().setHandlerChain(handlerList);

      // initialize ssl
      Map<String, Object> requestContext = ((BindingProvider) this.port).getRequestContext();
      
      if (webServiceURL.toLowerCase().startsWith("https")) {
      	log.trace("Using ssl for app2zuse request (using TrustAllManager - server certs may not be known).");
      	SSLContext sslContext = config.getSSLConfiguration().getSSLContext(true);
      	if (sslContext == null) {
					throw new Zuse2AppClientException("SSL context from configuration is empty. Please configure an SSL context in the configuration first.");
				}
				requestContext.put(JAXWSProperties.SSL_SOCKET_FACTORY, sslContext.getSocketFactory());
	     	log.trace("Using Lax hostname verfier for zuse2app communication (server certs not known).");
	     	requestContext.put(JAXWSProperties.HOSTNAME_VERIFIER, new LaxHostNameVerifier());
	    }
      log.trace("Sending request.");
      this.port.notification(request);
      
  	} catch(ZuseMsgException e) {
  		throw new Zuse2AppClientException(e);
  	} catch(Exception e) {
  		throw new Zuse2AppClientException(e);
  	}
  }

  public void sendDeliveryRequestEmail(String toAddress, DeliveryNotification deliveryNotification) throws Zuse2AppClientException {
  	
  	try {
	  	if (toAddress == null) {
	  		throw new NullPointerException("Argument 'toAddress' must not be null.");
	  	}
	  	
	  	log.debug("Sending zuse2app notification to application (email).");
	  	if (deliveryNotification == null ) {
	  		throw new NullPointerException("Argument 'request' must not be null.");
	  	}
	  	
	  	Zuse2AppConfiguration zuse2AppConfig = config.getZuse2AppConfiguration();
	  	if (zuse2AppConfig == null) {
	  		throw new Zuse2AppClientException("Cannot send notification. No zuse2app configuration found.");
	  	}
	  	
	    Properties properties = new Properties();
	    String mailHost = zuse2AppConfig.getZuse2AppMailHost();
	    if (mailHost == null) {
	    	throw new Zuse2AppClientException("No mail host in configuration found.");
	    }
	    log.trace("Mail host: " + mailHost);
	    properties.put("mail.smtp.host", mailHost);
	    
	    String mailPort = zuse2AppConfig.getZuse2AppMailPort();
	    log.trace("Mail port: " + mailPort == null ? "not defined" : mailPort);
	    if (zuse2AppConfig.getZuse2AppMailPort() != null) {
	    	properties.put("mail.smtp.port", zuse2AppConfig.getZuse2AppMailPort());
	    }
	    
	    Session session = Session.getDefaultInstance(properties, null );
	    MimeMessage message = new MimeMessage( session );
	    
	    String mailSenderName = zuse2AppConfig.getZuse2AppMailSenderName();
	    log.trace("Mail sender name: " + mailSenderName == null ? "not defined" : mailSenderName);
	    String mailSenderAddress = zuse2AppConfig.getZuse2AppMailSenderMailAddress();
	    if (mailSenderAddress == null) {
	    	throw new Zuse2AppClientException("No mail sender address in configuration found.");
	    }
	    log.trace("Mail sender address: " + mailSenderAddress);
	    
	    InternetAddress fromAddress = null;
	    if (mailSenderName != null) {
	    	fromAddress = new InternetAddress(mailSenderAddress, mailSenderName);
	    } else {
	    	fromAddress = new InternetAddress(mailSenderAddress);
	    }
	    message.setFrom(fromAddress);
	    log.trace("Setting to address: " + toAddress);
	    message.addRecipient(
	      Message.RecipientType.TO, new InternetAddress(toAddress)
	    );
	    
	    String subject = zuse2AppConfig.getZuse2AppMailSubject();
	    if (subject == null) {
	    	throw new Zuse2AppClientException("No mail subject in configuration found.");
	    }
	    log.trace("Mail subject: " + subject);
	    message.setSubject( subject );
	    Multipart mp = new MimeMultipart();
	
	    // string informational part
	    MailBody mailBody = zuse2AppConfig.getZuse2AppMailBody();
	    if (mailBody == null) {
	    	throw new Zuse2AppClientException("No mailbody found in configuration.");
	    }
	    ByteArrayOutputStream bos = new ByteArrayOutputStream();
	    MiscUtil.copyStream(mailBody.getInputStream(), bos);
	    log.trace("Mail body: " + bos.toString());
	    MimeBodyPart infoPart = new MimeBodyPart();
	    infoPart.setDataHandler(new DataHandler(mailBody));
	    mp.addBodyPart(infoPart);
	    
	    // serialize notification
	    bos = new ByteArrayOutputStream();
	    JAXBContext jaxbContext = JAXBContext.newInstance(DeliveryNotification.class.getPackage().getName());
	    jaxbContext.createMarshaller().marshal(deliveryNotification, bos);
	    log.trace("Notification content: " + bos.toString());
	    
	    // notification part
			// sign confirmation
	    log.trace("Signing delivery notification.");
			Zuse2AppSignatureCreator signatureCreator = new Zuse2AppSignatureCreator(this.config);
			Element signedNotificationElement = signatureCreator.createSignature(deliveryNotification);
			String signedNotification = DOMUtils.serializeNode(signedNotificationElement);
	    MimeBodyPart attachFilePart = new MimeBodyPart();
	    BinaryAttachment notificationAttachment = new BinaryAttachment("DeliveryNotification.xml", "text/xml", new ByteArrayInputStream(signedNotification.getBytes()));
	    attachFilePart.setDataHandler(new DataHandler(notificationAttachment));
	    attachFilePart.setFileName(notificationAttachment.getName());
	    mp.addBodyPart(attachFilePart);
	    message.setContent(mp);
	    Transport.send( message );
  	} catch(IOException e) {
  		log.warn("Error sending zuse2app mail.", e);
  	} catch (AddressException e) {
  		log.warn("Error sending zuse2app mail.", e);
    } catch (MessagingException e) {
    	log.warn("Error sending zuse2app mail.", e);
    } catch (JAXBException e) {
    	log.warn("Error sending zuse2app mail.", e);
    } catch (Zuse2AppSignatureException e) {
    	log.warn("Error sending zuse2app mail.", e);
    } catch (TransformerException e) {
    	log.warn("Error sending zuse2app mail.", e);
    }
  	
  }
  
}
