/*
 * Copyright 2011 E-Government Innovation Center (EGIZ)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package at.gv.zustellung.zusemsg.wsdl.client.handler;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;

import org.apache.commons.codec.binary.Base64;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Element;

import at.gv.util.MiscUtil;
import at.gv.util.client.moaspss.MOASPSSClient;
import at.gv.util.client.moaspss.MOASPSSClientException;
import at.gv.zustellung.zusemsg.config.ZuseMsgConfiguration;
import at.gv.zustellung.zusemsg.xsd.zuse.DeliveryNotification;

public class Zuse2AppSignatureCreator {

	private Logger log = LoggerFactory.getLogger(Zuse2AppSignatureCreator.class);
	private ZuseMsgConfiguration config = null;
	
	public Zuse2AppSignatureCreator(ZuseMsgConfiguration config) {
		if (config == null) {
			throw new NullPointerException("Argument 'config' must not be null.");
		}
		this.config = config;
	}
	
	public Element createSignature(DeliveryNotification deliveryNotification) throws Zuse2AppSignatureException {
    try {
			// sign message
	    // create moa-ss request
	  	log.trace("Reading MOA-SS template from resources.");
	    ByteArrayOutputStream bos = new ByteArrayOutputStream();
	    MiscUtil.copyStream(Zuse2AppSignatureCreationHandler.class.getResourceAsStream("/zuse2app.signature.creation.template.xml"), bos);
	    String template = bos.toString();
	    String moassServiceURL = this.config.getZuse2AppConfiguration().getZuse2AppMOASSServiceURL();
	    if (moassServiceURL == null) {
	    	log.warn("No MOA-SS service URL in configuration specified. Cannot sign message.");
	    	throw new Zuse2AppSignatureException("No MOA-SS service URL in configuration specified. Cannot sign message.");
	    }
	    log.trace("MOA-SS webService URL: " + moassServiceURL);
	    String keyIdentifier = this.config.getZuse2AppConfiguration().getZuse2AppMOASSKeyIdentifier();
	    if (keyIdentifier == null) {
	    	log.warn("No key identifier in configuration specified. Cannot sign message.");
	    	throw new Zuse2AppSignatureException("No key identifier in configuration specified. Cannot sign message.");
	    }
	    template = template.replaceAll("%1", keyIdentifier);
	    log.trace("MOA-SS Key Identifier: " + keyIdentifier);
	    
	    // save document as base64
	    // marshall
	    bos.reset();
	    JAXBContext ctx = JAXBContext.newInstance(DeliveryNotification.class.getPackage().getName());
	    ctx.createMarshaller().marshal(deliveryNotification, bos);
	    
	    String base64Document = new String(Base64.encodeBase64(bos.toByteArray()));
	    template = template.replaceAll("%2", base64Document);
	
	    // set signature location
	    String signatureLocation = null;
			if (deliveryNotification.getSuccess() != null) {
				signatureLocation = "/zs:DeliveryNotification/zs:Success";
			} else {
				signatureLocation = "/zs:DeliveryNotification/zs:Error";
			}
			template = template.replaceAll("%3", signatureLocation);
	  	
	    // call moa-ss
			log.trace("Calling signature creation service.");
			MOASPSSClient moaClient = new MOASPSSClient();
			return moaClient.sendSignatureCreationRequest(moassServiceURL, new ByteArrayInputStream(template.getBytes()));
    } catch(JAXBException e) {
    	throw new Zuse2AppSignatureException(e);
    } catch (IOException e) {
    	throw new Zuse2AppSignatureException(e);
    } catch (MOASPSSClientException e) {
    	throw new Zuse2AppSignatureException(e);
    }
	}
	
}
